"""
Describes problem and creates data structure.
This driver file is an example file to explain
how problems are created in the pyPR toolbox.

   Input:
       
           filename:  Name (string) of the file that contains the data.
                      The input is assumed to be of the type .mat
                      containing only one variable named 'pattern'.

    Output:
        
            input_data - A matlab structure containing various
                        information needed in the toolbox like
                        the problem type, physical parameters,
                        qualitative constraints and algorithmic
                        parameters.
For an explanation of the elements of the created structure,
we refer to the documentation.

Get preproccessed data. Right now, the method only returns the data
itself.
"""
from pyprProcessors import *
from pyprUtilities import *
import numpy as np
import matplotlib.pyplot as plt
from pyprAlgorithms import *

def simulatedDataCellImage():
    input_data = readDataFromFile('cell256.mat')

    # Since we don't know at this point which fields the readDataFromFile
    # function provides us, we check for neccessary information and provide
    # some reasonable choices ourselfes if they are not set yet.

    if 'problem' not in input_data:
        input_data["problemType"] = 'phase'     # 'phase', 'amplitude', 'mixed' # GGF HIER PROBLEM NAME, siehe unten    

    if 'algorithm' not in input_data:
        input_data["algorithm"] = {}
        input_data["algorithm"]["name"] = 'RAAR' # RAAR, MAP
        input_data["algorithm"]["error"] = 1e-12
        input_data["algorithm"]["maxit"] = 250
        input_data["algorithm"]["parameter"] = 'constant' # 'dynamic', 'constant'
        input_data["algorithm"]["beta"] = {}
        input_data["algorithm"]["beta"]["null"] = 0.99
        input_data["algorithm"]["beta"]["max"] = 0.55
        input_data["algorithm"]["beta"]["switch"] = 20
        input_data["algorithm"]["beta"]["value"] = 0.55
    
    if 'physicalParameters' not in input_data:
        input_data["physicalParameters"] = {}
        if input_data["problemType"] == 'phase':
            input_data["physicalParameters"]["delta"] = 1.6e-6
            input_data["physicalParameters"]["beta"] = 0.0
        elif input_data["problemType"] == 'amplitude':
            input_data["physicalParameters"]["delta"] = 0.0
            input_data["physicalParameters"]["beta"] = 8.0e-7
        else:                                               # mixed
            input_data["physicalParameters"]["delta"] = 1.6e-6
            input_data["physicalParameters"]["beta"] = 8.0e-7


        input_data["physicalParameters"]["tau"] = 1e-5
        input_data["physicalParameters"]["lambda"] = 1e-10
        input_data["physicalParameters"]["k"] = 2*np.pi / input_data["physicalParameters"]["lambda"]
        input_data["physicalParameters"]["P"] = input_data["data"] * input_data["physicalParameters"]["tau"]
        input_data["physicalParameters"]["A"] = input_data["data"] * input_data["physicalParameters"]["tau"]
        input_data["Psi"] = np.exp(-1j * input_data["physicalParameters"]["k"]
                                    * input_data["physicalParameters"]["delta"]
                                    * input_data["physicalParameters"]["P"]
                                    - input_data["physicalParameters"]["k"]
                                    * input_data["physicalParameters"]["beta"]
                                    * input_data["physicalParameters"]["A"]
                                )
    if 'noise' not in input_data:
        input_data["noise"] = {}
        input_data["noise"]["type"] = 'poisson' 
        input_data["noise"]["level"] = 6.75    
        input_data["noise"]["error"] = 1e-2     # describes the phattening of M
        
    if 'problem' not in input_data:
        input_data["problem"] = {}
        input_data["problem"]["propagator"] = 'Fresnel' # 'Fresnel', 'Fourier'
        input_data["problem"]["Fx"] = 4e-3
        input_data["problem"]["Fy"] = 4e-3
        input_data["problem"]["data"] = wavePropagator(input_data, input_data["Psi"], 1)
        input_data["problem"]["data"] = np.abs(wavePropagator(input_data, input_data["Psi"], 1))
        input_data["problem"]["data"] = createNoisyMeasurement(input_data)
        input_data["problem"]["constraint"] = {}
        input_data["problem"]["constraint"]["name"]      = 'shearlet thresholding'                      # shearlet, support, negativity, amplitude
        input_data["problem"]["constraint"]["info"]      = 'smooth hard'                # optional (depends on constraint name). soft / smooth hard
        input_data["problem"]["constraint"]["parameter"] = 'constant'                   # dynamic = 1/k right now, const
        input_data["problem"]["constraint"]["gamma0"]    = 0.007                       # initial threshold parameter
        input_data["problem"]["constraint"]["support"]   = createSupportIdx(input_data, 0.8, 'ball') # for the cell and box, we would chose R=0.75.
    return input_data